@@ -1,3 +1,18 @@
+/*
+ * Copyright 2014 - Present Rafael Winterhalter
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
 package net.bytebuddy.android;
 
 import android.annotation.TargetApi;
@@ -7,6 +22,7 @@ import com.android.dx.cf.direct.StdAttributeFactory;
 import com.android.dx.dex.DexOptions;
 import com.android.dx.dex.cf.CfOptions;
 import com.android.dx.dex.cf.CfTranslator;
+import com.android.dx.dex.file.ClassDefItem;
 import com.android.dx.dex.file.DexFile;
 import dalvik.system.BaseDexClassLoader;
 import dalvik.system.DexClassLoader;
@@ -14,10 +30,19 @@ import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
 import net.bytebuddy.description.type.TypeDescription;
 import net.bytebuddy.dynamic.loading.ClassLoadingStrategy;
 import net.bytebuddy.utility.RandomString;
+import net.bytebuddy.utility.nullability.AlwaysNull;
+import net.bytebuddy.utility.nullability.MaybeNull;
 
-import java.io.*;
+import java.io.File;
+import java.io.FileOutputStream;
+import java.io.IOException;
+import java.io.OutputStream;
+import java.io.Writer;
+import java.lang.reflect.Constructor;
+import java.lang.reflect.Field;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
+import java.util.Collections;
 import java.util.HashMap;
 import java.util.Map;
 import java.util.Set;
@@ -35,8 +60,8 @@ import java.util.logging.Logger;
  * <b>not</b> permitted by Android's security checks to store these files in a shared folder where they could be
  * manipulated by a third application what would break Android's sandbox model. An example for a forbidden storage
  * would therefore be the external storage. Instead, the class loading application must either supply a designated
- * directory, such as by creating a directory using {@link android.content.Context#getDir(String, int)} with specifying
- * {@link android.content.Context#MODE_PRIVATE} visibility for the created folder or by using the
+ * directory, such as by creating a directory using {@code android.content.Context#getDir(String, int)} with specifying
+ * {@code android.content.Context#MODE_PRIVATE} visibility for the created folder or by using the
  * {@code getCodeCacheDir} directory which is exposed for Android API versions 21 or higher.
  * </p>
  * <p>
@@ -68,8 +93,34 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
     /**
      * A value for a {@link dalvik.system.DexClassLoader} to indicate that the library path is empty.
      */
+    @AlwaysNull
     private static final String EMPTY_LIBRARY_PATH = null;
 
+    /**
+     * A processor for files before adding them to a dex file.
+     */
+    private static final FileProcessor FILE_PROCESSOR;
+
+    /*
+     * Resolves the file processor.
+     */
+    static {
+        FileProcessor fileProcessor;
+        try {
+            fileProcessor = new FileProcessor.ForReadOnlyClassFile(
+                    Class.forName("java.nio.file.Files").getMethod("setPosixFilePermissions",
+                            Class.forName("java.nio.file.Path"),
+                            Set.class),
+                    File.class.getMethod("toPath"),
+                    Collections.singleton(Class.forName("java.nio.file.attribute.PosixFilePermission")
+                            .getMethod("valueOf", String.class)
+                            .invoke(null, "OWNER_READ")));
+        } catch (Throwable ignored) {
+            fileProcessor = FileProcessor.Disabled.INSTANCE;
+        }
+        FILE_PROCESSOR = fileProcessor;
+    }
+
     /**
      * The dex creator to be used by this Android class loading strategy.
      */
@@ -103,8 +154,10 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
         randomString = new RandomString();
     }
 
-    @Override
-    public Map<TypeDescription, Class<?>> load(ClassLoader classLoader, Map<TypeDescription, byte[]> types) {
+    /**
+     * {@inheritDoc}
+     */
+    public Map<TypeDescription, Class<?>> load(@MaybeNull ClassLoader classLoader, Map<TypeDescription, byte[]> types) {
         DexProcessor.Conversion conversion = dexProcessor.create();
         for (Map.Entry<TypeDescription, byte[]> entry : types.entrySet()) {
             conversion.register(entry.getKey().getName(), entry.getValue());
@@ -114,14 +167,17 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
             if (!jar.createNewFile()) {
                 throw new IllegalStateException("Cannot create " + jar);
             }
-            JarOutputStream zipOutputStream = new JarOutputStream(new FileOutputStream(jar));
+            OutputStream outputStream = new FileOutputStream(jar);
             try {
-                zipOutputStream.putNextEntry(new JarEntry(DEX_CLASS_FILE));
-                conversion.drainTo(zipOutputStream);
-                zipOutputStream.closeEntry();
+                JarOutputStream jarOutputStream = new JarOutputStream(outputStream);
+                jarOutputStream.putNextEntry(new JarEntry(DEX_CLASS_FILE));
+                conversion.drainTo(jarOutputStream);
+                jarOutputStream.closeEntry();
+                jarOutputStream.close();
             } finally {
-                zipOutputStream.close();
+                outputStream.close();
             }
+            FILE_PROCESSOR.accept(jar);
             return doLoad(classLoader, types.keySet(), jar);
         } catch (IOException exception) {
             throw new IllegalStateException("Cannot write to zip file " + jar, exception);
@@ -141,7 +197,87 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
      * @return A mapping of all type descriptions to their loaded types.
      * @throws IOException If an I/O exception occurs.
      */
-    protected abstract Map<TypeDescription, Class<?>> doLoad(ClassLoader classLoader, Set<TypeDescription> typeDescriptions, File jar) throws IOException;
+    protected abstract Map<TypeDescription, Class<?>> doLoad(@MaybeNull ClassLoader classLoader, Set<TypeDescription> typeDescriptions, File jar) throws IOException;
+
+    /**
+     * A processor for files that are added to a dex file.
+     */
+    protected interface FileProcessor {
+
+        /**
+         * Accepts a file for processing.
+         *
+         * @param file The file to process.
+         */
+        void accept(File file);
+
+        /**
+         * A non-operational file processor.
+         */
+        enum Disabled implements FileProcessor {
+
+            /**
+             * The singleton instance.
+             */
+            INSTANCE;
+
+            /**
+             * {@inheritDoc}
+             */
+            public void accept(File file) {
+                /* do nothing */
+            }
+        }
+
+        /**
+         * A file processor that marks a file as read-only.
+         */
+        class ForReadOnlyClassFile implements FileProcessor {
+
+            /**
+             * The {@code java.nio.file.Files#setPosixFilePermissions} method.
+             */
+            private final Method setPosixFilePermissions;
+
+            /**
+             * The {@code java.io.File#toPath} method.
+             */
+            private final Method toPath;
+
+            /**
+             * The set of permissions to set.
+             */
+            private final Set<?> permissions;
+
+            /**
+             * Creates a new file processor for a read only file.
+             *
+             * @param setPosixFilePermissions The {@code java.nio.file.Files#setPosixFilePermissions} method.
+             * @param toPath                  The {@code java.io.File#toPath} method.
+             * @param permissions             The set of {java.nio.file.attribute.FilePermission} to set.
+             */
+            protected ForReadOnlyClassFile(Method setPosixFilePermissions, Method toPath, Set<?> permissions) {
+                this.setPosixFilePermissions = setPosixFilePermissions;
+                this.toPath = toPath;
+                this.permissions = permissions;
+            }
+
+            /**
+             * {@inheritDoc}
+             */
+            public void accept(File file) {
+                try {
+                    setPosixFilePermissions.invoke(null, toPath.invoke(file), permissions);
+                } catch (IllegalAccessException exception) {
+                    throw new IllegalStateException("Cannot access file system permissions", exception);
+                } catch (InvocationTargetException exception) {
+                    if (!(exception.getTargetException() instanceof UnsupportedOperationException)) {
+                        throw new IllegalStateException("Cannot invoke file system permissions method", exception.getTargetException());
+                    }
+                }
+            }
+        }
+    }
 
     /**
      * A dex processor is responsible for converting a collection of Java class files into a Android dex file.
@@ -189,6 +325,49 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
              */
             private static final int DEX_COMPATIBLE_API_VERSION = 13;
 
+            /**
+             * The dispatcher for translating a dex file.
+             */
+            private static final Dispatcher DISPATCHER;
+
+            /*
+             * Resolves the dispatcher for class file translations.
+             */
+            static {
+                Dispatcher dispatcher;
+                try {
+                    Class<?> dxContextType = Class.forName("com.android.dx.command.dexer.DxContext");
+                    dispatcher = new Dispatcher.ForApi26LevelCompatibleVm(CfTranslator.class.getMethod("translate",
+                            dxContextType,
+                            DirectClassFile.class,
+                            byte[].class,
+                            CfOptions.class,
+                            DexOptions.class,
+                            DexFile.class), dxContextType.getConstructor());
+                } catch (Throwable ignored) {
+                    try {
+                        dispatcher = new Dispatcher.ForLegacyVm(CfTranslator.class.getMethod("translate",
+                                DirectClassFile.class,
+                                byte[].class,
+                                CfOptions.class,
+                                DexOptions.class,
+                                DexFile.class), DexOptions.class.getField("minSdkVersion"));
+                    } catch (Throwable suppressed) {
+                        try {
+                            dispatcher = new Dispatcher.ForLegacyVm(CfTranslator.class.getMethod("translate",
+                                    DirectClassFile.class,
+                                    byte[].class,
+                                    CfOptions.class,
+                                    DexOptions.class,
+                                    DexFile.class), DexOptions.class.getField("targetApiLevel"));
+                        } catch (Throwable throwable) {
+                            dispatcher = new Dispatcher.Unavailable(throwable.getMessage());
+                        }
+                    }
+                }
+                DISPATCHER = dispatcher;
+            }
+
             /**
              * Creates a default dex processor that ensures API version compatibility.
              *
@@ -196,7 +375,7 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
              */
             protected static DexProcessor makeDefault() {
                 DexOptions dexOptions = new DexOptions();
-                dexOptions.targetApiLevel = DEX_COMPATIBLE_API_VERSION;
+                DISPATCHER.setTargetApi(dexOptions, DEX_COMPATIBLE_API_VERSION);
                 return new ForSdkCompiler(dexOptions, new CfOptions());
             }
 
@@ -208,6 +387,7 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
             /**
              * Indicates that a dex file should be written without providing a human readable output.
              */
+            @AlwaysNull
             private static final Writer NO_PRINT_OUTPUT = null;
 
             /**
@@ -236,7 +416,9 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
                 this.dexCompilerOptions = dexCompilerOptions;
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public DexProcessor.Conversion create() {
                 return new Conversion(new DexFile(dexFileOptions));
             }
@@ -266,41 +448,213 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
                     this.dexFile = dexFile;
                 }
 
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
                 public void register(String name, byte[] binaryRepresentation) {
-                    DirectClassFile directClassFile = new DirectClassFile(binaryRepresentation, name.replace('.', '/') + CLASS_FILE_EXTENSION, NON_STRICT);
+                    DirectClassFile directClassFile = new DirectClassFile(binaryRepresentation,
+                            name.replace('.', '/') + CLASS_FILE_EXTENSION,
+                            NON_STRICT);
                     directClassFile.setAttributeFactory(new StdAttributeFactory());
-                    dexFile.add(CfTranslator.translate(directClassFile,
+                    dexFile.add(DISPATCHER.translate(directClassFile,
                             binaryRepresentation,
                             dexCompilerOptions,
                             dexFileOptions,
                             new DexFile(dexFileOptions)));
                 }
 
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
                 public void drainTo(OutputStream outputStream) throws IOException {
                     dexFile.writeTo(outputStream, NO_PRINT_OUTPUT, NOT_VERBOSE);
                 }
+            }
+
+            /**
+             * A dispatcher for translating a direct class file.
+             */
+            protected interface Dispatcher {
 
                 /**
-                 * Returns the outer instance.
+                 * Creates a new class file definition.
                  *
-                 * @return The outer instance.
+                 * @param directClassFile      The direct class file to translate.
+                 * @param binaryRepresentation The file's binary representation.
+                 * @param dexCompilerOptions   The dex compiler options.
+                 * @param dexFileOptions       The dex file options.
+                 * @param dexFile              The dex file.
+                 * @return The translated class file definition.
                  */
-                private ForSdkCompiler getOuter() {
-                    return ForSdkCompiler.this;
+                ClassDefItem translate(DirectClassFile directClassFile,
+                                       byte[] binaryRepresentation,
+                                       CfOptions dexCompilerOptions,
+                                       DexOptions dexFileOptions,
+                                       DexFile dexFile);
+
+                /**
+                 * Sets the target API level if available.
+                 *
+                 * @param dexOptions     The dex options to set the api version for
+                 * @param targetApiLevel The target API level.
+                 */
+                void setTargetApi(DexOptions dexOptions, int targetApiLevel);
+
+                /**
+                 * An unavailable dispatcher.
+                 */
+                class Unavailable implements Dispatcher {
+
+                    /**
+                     * A message explaining why the dispatcher is unavailable.
+                     */
+                    private final String message;
+
+                    /**
+                     * Creates a new unavailable dispatcher.
+                     *
+                     * @param message A message explaining why the dispatcher is unavailable.
+                     */
+                    protected Unavailable(String message) {
+                        this.message = message;
+                    }
+
+                    /**
+                     * {@inheritDoc}
+                     */
+                    public ClassDefItem translate(DirectClassFile directClassFile,
+                                                  byte[] binaryRepresentation,
+                                                  CfOptions dexCompilerOptions,
+                                                  DexOptions dexFileOptions,
+                                                  DexFile dexFile) {
+                        throw new IllegalStateException("Could not resolve dispatcher: " + message);
+                    }
+
+                    /**
+                     * {@inheritDoc}
+                     */
+                    public void setTargetApi(DexOptions dexOptions, int targetApiLevel) {
+                        throw new IllegalStateException("Could not resolve dispatcher: " + message);
+                    }
                 }
 
-                @Override // HE: Remove when Lombok support for getOuter is added.
-                public boolean equals(Object other) {
-                    return this == other || !(other == null || getClass() != other.getClass())
-                            && ForSdkCompiler.this.equals(((Conversion) other).getOuter())
-                            && dexFile.equals(((Conversion) other).dexFile);
+                /**
+                 * A dispatcher for a lagacy Android VM.
+                 */
+                class ForLegacyVm implements Dispatcher {
+
+                    /**
+                     * The {@code CfTranslator#translate(DirectClassFile, byte[], CfOptions, DexOptions, DexFile)} method.
+                     */
+                    private final Method translate;
+
+                    /**
+                     * The {@code DexOptions#targetApiLevel} field.
+                     */
+                    private final Field targetApi;
+
+                    /**
+                     * Creates a new dispatcher.
+                     *
+                     * @param translate The {@code CfTranslator#translate(DirectClassFile, byte[], CfOptions, DexOptions, DexFile)} method.
+                     * @param targetApi The {@code DexOptions#targetApiLevel} field.
+                     */
+                    protected ForLegacyVm(Method translate, Field targetApi) {
+                        this.translate = translate;
+                        this.targetApi = targetApi;
+                    }
+
+                    /**
+                     * {@inheritDoc}
+                     */
+                    public ClassDefItem translate(DirectClassFile directClassFile,
+                                                  byte[] binaryRepresentation,
+                                                  CfOptions dexCompilerOptions,
+                                                  DexOptions dexFileOptions,
+                                                  DexFile dexFile) {
+                        try {
+                            return (ClassDefItem) translate.invoke(null,
+                                    directClassFile,
+                                    binaryRepresentation,
+                                    dexCompilerOptions,
+                                    dexFileOptions,
+                                    dexFile);
+                        } catch (IllegalAccessException exception) {
+                            throw new IllegalStateException("Cannot access an Android dex file translation method", exception);
+                        } catch (InvocationTargetException exception) {
+                            throw new IllegalStateException("Cannot invoke Android dex file translation method", exception.getTargetException());
+                        }
+                    }
+
+                    /**
+                     * {@inheritDoc}
+                     */
+                    public void setTargetApi(DexOptions dexOptions, int targetApiLevel) {
+                        try {
+                            targetApi.set(dexOptions, targetApiLevel);
+                        } catch (IllegalAccessException exception) {
+                            throw new IllegalStateException("Cannot access an Android dex file translation method", exception);
+                        }
+                    }
                 }
 
-                @Override // HE: Remove when Lombok support for getOuter is added.
-                public int hashCode() {
-                    return dexFile.hashCode() + 31 * ForSdkCompiler.this.hashCode();
+                /**
+                 * A dispatcher for an Android VM with API level 26 or higher.
+                 */
+                class ForApi26LevelCompatibleVm implements Dispatcher {
+
+                    /**
+                     * The {@code CfTranslator#translate(DxContext, DirectClassFile, byte[], CfOptions, DexOptions, DexFile)} method.
+                     */
+                    private final Method translate;
+
+                    /**
+                     * The {@code com.android.dx.command.dexer.DxContext#DxContext()} constructor.
+                     */
+                    private final Constructor<?> dxContext;
+
+                    /**
+                     * Creates a new dispatcher.
+                     *
+                     * @param translate The {@code CfTranslator#translate(DxContext, DirectClassFile, byte[], CfOptions, DexOptions, DexFile)} method.
+                     * @param dxContext The {@code com.android.dx.command.dexer.DxContext#DxContext()} constructor.
+                     */
+                    protected ForApi26LevelCompatibleVm(Method translate, Constructor<?> dxContext) {
+                        this.translate = translate;
+                        this.dxContext = dxContext;
+                    }
+
+                    /**
+                     * {@inheritDoc}
+                     */
+                    public ClassDefItem translate(DirectClassFile directClassFile,
+                                                  byte[] binaryRepresentation,
+                                                  CfOptions dexCompilerOptions,
+                                                  DexOptions dexFileOptions,
+                                                  DexFile dexFile) {
+                        try {
+                            return (ClassDefItem) translate.invoke(null,
+                                    dxContext.newInstance(),
+                                    directClassFile,
+                                    binaryRepresentation,
+                                    dexCompilerOptions,
+                                    dexFileOptions,
+                                    dexFile);
+                        } catch (IllegalAccessException exception) {
+                            throw new IllegalStateException("Cannot access an Android dex file translation method", exception);
+                        } catch (InstantiationException exception) {
+                            throw new IllegalStateException("Cannot instantiate dex context", exception);
+                        } catch (InvocationTargetException exception) {
+                            throw new IllegalStateException("Cannot invoke Android dex file translation method", exception.getTargetException());
+                        }
+                    }
+
+                    /**
+                     * {@inheritDoc}
+                     */
+                    public void setTargetApi(DexOptions dexOptions, int targetApiLevel) {
+                        /* do nothing */
+                    }
                 }
             }
         }
@@ -333,9 +687,11 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
             super(privateDirectory, dexProcessor);
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         @SuppressFBWarnings(value = "DP_CREATE_CLASSLOADER_INSIDE_DO_PRIVILEGED", justification = "Android discourages the use of access controllers")
-        protected Map<TypeDescription, Class<?>> doLoad(ClassLoader classLoader, Set<TypeDescription> typeDescriptions, File jar) {
+        protected Map<TypeDescription, Class<?>> doLoad(@MaybeNull ClassLoader classLoader, Set<TypeDescription> typeDescriptions, File jar) {
             ClassLoader dexClassLoader = new DexClassLoader(jar.getAbsolutePath(), privateDirectory.getAbsolutePath(), EMPTY_LIBRARY_PATH, classLoader);
             Map<TypeDescription, Class<?>> loadedTypes = new HashMap<TypeDescription, Class<?>>();
             for (TypeDescription typeDescription : typeDescriptions) {
@@ -394,28 +750,38 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
             super(privateDirectory, dexProcessor);
         }
 
-        @Override
-        public Map<TypeDescription, Class<?>> load(ClassLoader classLoader, Map<TypeDescription, byte[]> types) {
+        /**
+         * {@inheritDoc}
+         */
+        public Map<TypeDescription, Class<?>> load(@MaybeNull ClassLoader classLoader, Map<TypeDescription, byte[]> types) {
             if (classLoader == null) {
                 throw new IllegalArgumentException("Cannot inject classes into the bootstrap class loader on Android");
             }
             return super.load(classLoader, types);
         }
 
-        @Override
-        protected Map<TypeDescription, Class<?>> doLoad(ClassLoader classLoader, Set<TypeDescription> typeDescriptions, File jar) throws IOException {
+        /**
+         * {@inheritDoc}
+         */
+        protected Map<TypeDescription, Class<?>> doLoad(@MaybeNull ClassLoader classLoader, Set<TypeDescription> typeDescriptions, File jar) throws IOException {
             dalvik.system.DexFile dexFile = DISPATCHER.loadDex(privateDirectory, jar, classLoader, randomString);
-            Map<TypeDescription, Class<?>> loadedTypes = new HashMap<TypeDescription, Class<?>>();
-            for (TypeDescription typeDescription : typeDescriptions) {
-                synchronized (classLoader) { // Guaranteed to be non-null by check in 'load' method.
-                    Class<?> type = DISPATCHER.loadClass(dexFile, classLoader, typeDescription);
-                    if (type == null) {
-                        throw new IllegalStateException("Could not load " + typeDescription);
+            try {
+                Map<TypeDescription, Class<?>> loadedTypes = new HashMap<TypeDescription, Class<?>>();
+                for (TypeDescription typeDescription : typeDescriptions) {
+                    synchronized (classLoader) { // Guaranteed to be non-null by check in 'load' method.
+                        Class<?> type = DISPATCHER.loadClass(dexFile, classLoader, typeDescription);
+                        if (type == null) {
+                            throw new IllegalStateException("Could not load " + typeDescription);
+                        }
+                        loadedTypes.put(typeDescription, type);
                     }
-                    loadedTypes.put(typeDescription, type);
+                }
+                return loadedTypes;
+            } finally {
+                if (dexFile != null) {
+                    dexFile.close();
                 }
             }
-            return loadedTypes;
         }
 
         /**
@@ -433,7 +799,8 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
              * @return The created {@link dalvik.system.DexFile} or {@code null} if no such file is created.
              * @throws IOException If an I/O exception is thrown.
              */
-            dalvik.system.DexFile loadDex(File privateDirectory, File jar, ClassLoader classLoader, RandomString randomString) throws IOException;
+            @MaybeNull
+            dalvik.system.DexFile loadDex(File privateDirectory, File jar, @MaybeNull ClassLoader classLoader, RandomString randomString) throws IOException;
 
             /**
              * Loads a class.
@@ -443,7 +810,8 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
              * @param typeDescription The type to load.
              * @return The loaded class.
              */
-            Class<?> loadClass(dalvik.system.DexFile dexFile, ClassLoader classLoader, TypeDescription typeDescription);
+            @MaybeNull
+            Class<?> loadClass(dalvik.system.DexFile dexFile, @MaybeNull ClassLoader classLoader, TypeDescription typeDescription);
 
             /**
              * A dispatcher for legacy VMs that allow {@link dalvik.system.DexFile#loadDex(String, String, int)}.
@@ -465,18 +833,24 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
                  */
                 private static final String EXTENSION = ".data";
 
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
+                @MaybeNull
                 public dalvik.system.DexFile loadDex(File privateDirectory,
                                                      File jar,
-                                                     ClassLoader classLoader,
+                                                     @MaybeNull ClassLoader classLoader,
                                                      RandomString randomString) throws IOException {
                     return dalvik.system.DexFile.loadDex(jar.getAbsolutePath(),
                             new File(privateDirectory.getAbsolutePath(), randomString.nextString() + EXTENSION).getAbsolutePath(),
                             NO_FLAGS);
                 }
 
-                @Override
-                public Class<?> loadClass(dalvik.system.DexFile dexFile, ClassLoader classLoader, TypeDescription typeDescription) {
+                /**
+                 * {@inheritDoc}
+                 */
+                @MaybeNull
+                public Class<?> loadClass(dalvik.system.DexFile dexFile, @MaybeNull ClassLoader classLoader, TypeDescription typeDescription) {
                     return dexFile.loadClass(typeDescription.getName(), classLoader);
                 }
             }
@@ -489,6 +863,7 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
                 /**
                  * Indicates that this dispatcher does not return a {@link dalvik.system.DexFile} instance.
                  */
+                @AlwaysNull
                 private static final dalvik.system.DexFile NO_RETURN_VALUE = null;
 
                 /**
@@ -505,21 +880,24 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
                     this.addDexPath = addDexPath;
                 }
 
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
+                @MaybeNull
                 public dalvik.system.DexFile loadDex(File privateDirectory,
                                                      File jar,
-                                                     ClassLoader classLoader,
+                                                     @MaybeNull ClassLoader classLoader,
                                                      RandomString randomString) throws IOException {
                     if (!(classLoader instanceof BaseDexClassLoader)) {
                         throw new IllegalArgumentException("On Android P, a class injection can only be applied to BaseDexClassLoader: " + classLoader);
                     }
                     try {
-                        addDexPath.invoke(classLoader, jar.getAbsolutePath(), true);
+                        addDexPath.invoke(classLoader, jar.getAbsolutePath(), false);
                         return NO_RETURN_VALUE;
                     } catch (IllegalAccessException exception) {
                         throw new IllegalStateException("Cannot access BaseDexClassLoader#addDexPath(String, boolean)", exception);
                     } catch (InvocationTargetException exception) {
-                        Throwable cause = exception.getCause();
+                        Throwable cause = exception.getTargetException();
                         if (cause instanceof IOException) {
                             throw (IOException) cause;
                         } else {
@@ -528,8 +906,10 @@ public abstract class AndroidClassLoadingStrategy implements ClassLoadingStrateg
                     }
                 }
 
-                @Override
-                public Class<?> loadClass(dalvik.system.DexFile dexFile, ClassLoader classLoader, TypeDescription typeDescription) {
+                /**
+                 * {@inheritDoc}
+                 */
+                public Class<?> loadClass(@MaybeNull dalvik.system.DexFile dexFile, @MaybeNull ClassLoader classLoader, TypeDescription typeDescription) {
                     try {
                         return Class.forName(typeDescription.getName(), false, classLoader);
                     } catch (ClassNotFoundException exception) {
