@@ -1,6 +1,21 @@
+/*
+ * Copyright 2014 - Present Rafael Winterhalter
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
 package net.bytebuddy.implementation;
 
-import net.bytebuddy.ClassFileVersion;
+import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
 import net.bytebuddy.build.HashCodeAndEqualsPlugin;
 import net.bytebuddy.description.field.FieldDescription;
 import net.bytebuddy.description.method.MethodDescription;
@@ -8,6 +23,7 @@ import net.bytebuddy.description.type.TypeDefinition;
 import net.bytebuddy.description.type.TypeDescription;
 import net.bytebuddy.dynamic.scaffold.InstrumentedType;
 import net.bytebuddy.implementation.bytecode.*;
+import net.bytebuddy.implementation.bytecode.constant.ClassConstant;
 import net.bytebuddy.implementation.bytecode.constant.IntegerConstant;
 import net.bytebuddy.implementation.bytecode.member.FieldAccess;
 import net.bytebuddy.implementation.bytecode.member.MethodInvocation;
@@ -19,6 +35,7 @@ import org.objectweb.asm.MethodVisitor;
 import org.objectweb.asm.Opcodes;
 
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.List;
 
 import static net.bytebuddy.matcher.ElementMatchers.*;
@@ -49,6 +66,14 @@ public class HashCodeMethod implements Implementation {
             .filter(isHashCode())
             .getOnly();
 
+    /**
+     * The {@link Object#getClass()} method.
+     */
+    private static final MethodDescription.InDefinedShape GET_CLASS = TypeDescription.ForLoadedType.of(Object.class)
+            .getDeclaredMethods()
+            .filter(named("getClass").and(takesArguments(0)))
+            .getOnly();
+
     /**
      * The hash code's offset provider.
      */
@@ -69,13 +94,18 @@ public class HashCodeMethod implements Implementation {
      */
     private final ElementMatcher.Junction<? super FieldDescription.InDefinedShape> nonNullable;
 
+    /**
+     * A matcher to determine that a field should be considered by its identity.
+     */
+    private final ElementMatcher.Junction<? super FieldDescription.InDefinedShape> identity;
+
     /**
      * Creates a new hash code method implementation.
      *
      * @param offsetProvider The hash code's offset provider.
      */
     protected HashCodeMethod(OffsetProvider offsetProvider) {
-        this(offsetProvider, DEFAULT_MULTIPLIER, none(), none());
+        this(offsetProvider, DEFAULT_MULTIPLIER, none(), none(), none());
     }
 
     /**
@@ -85,15 +115,18 @@ public class HashCodeMethod implements Implementation {
      * @param multiplier     A multiplier for each value before adding a field's hash code value
      * @param ignored        A matcher to filter fields that should not be used for a hash codes computation.
      * @param nonNullable    A matcher to determine fields of a reference type that cannot be {@code null}.
+     * @param identity       A matcher to determine that a field should be considered by its identity.
      */
     private HashCodeMethod(OffsetProvider offsetProvider,
                            int multiplier,
                            ElementMatcher.Junction<? super FieldDescription.InDefinedShape> ignored,
-                           ElementMatcher.Junction<? super FieldDescription.InDefinedShape> nonNullable) {
+                           ElementMatcher.Junction<? super FieldDescription.InDefinedShape> nonNullable,
+                           ElementMatcher.Junction<? super FieldDescription.InDefinedShape> identity) {
         this.offsetProvider = offsetProvider;
         this.multiplier = multiplier;
         this.ignored = ignored;
         this.nonNullable = nonNullable;
+        this.identity = identity;
     }
 
     /**
@@ -105,6 +138,16 @@ public class HashCodeMethod implements Implementation {
         return new HashCodeMethod(OffsetProvider.ForSuperMethodCall.INSTANCE);
     }
 
+    /**
+     * Creates a hash code method implementation that bases the hash code on the instrumented type's class constant's hash code..
+     *
+     * @param dynamic {@code true} if the type should be resolved from the instance and not be set as the declaring class.
+     * @return A hash code method implementation that bases the hash code on the instrumented type's class constant's hash code.
+     */
+    public static HashCodeMethod usingTypeHashOffset(boolean dynamic) {
+        return new HashCodeMethod(dynamic ? OffsetProvider.ForDynamicTypeHash.INSTANCE : OffsetProvider.ForStaticTypeHash.INSTANCE);
+    }
+
     /**
      * Creates a hash code method implementation that bases the hash code on a fixed value.
      *
@@ -132,7 +175,7 @@ public class HashCodeMethod implements Implementation {
      * @return A new version of this hash code method implementation that also ignores any fields matched by the provided matcher.
      */
     public HashCodeMethod withIgnoredFields(ElementMatcher<? super FieldDescription.InDefinedShape> ignored) {
-        return new HashCodeMethod(offsetProvider, multiplier, this.ignored.<FieldDescription.InDefinedShape>or(ignored), nonNullable);
+        return new HashCodeMethod(offsetProvider, multiplier, this.ignored.<FieldDescription.InDefinedShape>or(ignored), nonNullable, identity);
     }
 
     /**
@@ -144,7 +187,17 @@ public class HashCodeMethod implements Implementation {
      * the provided matcher.
      */
     public HashCodeMethod withNonNullableFields(ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable) {
-        return new HashCodeMethod(offsetProvider, multiplier, ignored, this.nonNullable.<FieldDescription.InDefinedShape>or(nonNullable));
+        return new HashCodeMethod(offsetProvider, multiplier, ignored, this.nonNullable.<FieldDescription.InDefinedShape>or(nonNullable), identity);
+    }
+
+    /**
+     * Returns a new version of this hash code method implementation that considers the matched fields by their identity.
+     *
+     * @param identity A matcher to specify any fields that should be considered by their identity.
+     * @return A new version of this hash code method implementation that also considers the matched fields by their identity.
+     */
+    public HashCodeMethod withIdentityFields(ElementMatcher<? super FieldDescription.InDefinedShape> identity) {
+        return new HashCodeMethod(offsetProvider, multiplier, ignored, nonNullable, this.identity.<FieldDescription.InDefinedShape>or(identity));
     }
 
     /**
@@ -159,15 +212,19 @@ public class HashCodeMethod implements Implementation {
         if (multiplier == 0) {
             throw new IllegalArgumentException("Hash code multiplier must not be zero");
         }
-        return new HashCodeMethod(offsetProvider, multiplier, ignored, nonNullable);
+        return new HashCodeMethod(offsetProvider, multiplier, ignored, nonNullable, identity);
     }
 
-    @Override
+    /**
+     * {@inheritDoc}
+     */
     public InstrumentedType prepare(InstrumentedType instrumentedType) {
         return instrumentedType;
     }
 
-    @Override
+    /**
+     * {@inheritDoc}
+     */
     public ByteCodeAppender appender(Target implementationTarget) {
         if (implementationTarget.getInstrumentedType().isInterface()) {
             throw new IllegalStateException("Cannot implement meaningful hash code method for " + implementationTarget.getInstrumentedType());
@@ -175,7 +232,8 @@ public class HashCodeMethod implements Implementation {
         return new Appender(offsetProvider.resolve(implementationTarget.getInstrumentedType()),
                 multiplier,
                 implementationTarget.getInstrumentedType().getDeclaredFields().filter(not(isStatic().or(ignored))),
-                nonNullable);
+                nonNullable,
+                identity);
     }
 
     /**
@@ -211,7 +269,9 @@ public class HashCodeMethod implements Implementation {
                 this.value = value;
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation resolve(TypeDescription instrumentedType) {
                 return IntegerConstant.forValue(value);
             }
@@ -227,7 +287,9 @@ public class HashCodeMethod implements Implementation {
              */
             INSTANCE;
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation resolve(TypeDescription instrumentedType) {
                 TypeDefinition superClass = instrumentedType.getSuperClass();
                 if (superClass == null) {
@@ -236,6 +298,44 @@ public class HashCodeMethod implements Implementation {
                 return new StackManipulation.Compound(MethodVariableAccess.loadThis(), MethodInvocation.invoke(HASH_CODE).special(superClass.asErasure()));
             }
         }
+
+        /**
+         * An offset provider that uses the instrumented type's class constant's hash code.
+         */
+        enum ForStaticTypeHash implements OffsetProvider {
+
+            /**
+             * The singleton instance.
+             */
+            INSTANCE;
+
+            /**
+             * {@inheritDoc}
+             */
+            public StackManipulation resolve(TypeDescription instrumentedType) {
+                return new StackManipulation.Compound(ClassConstant.of(instrumentedType), MethodInvocation.invoke(HASH_CODE).virtual(TypeDescription.ForLoadedType.of(Class.class)));
+            }
+        }
+
+        /**
+         * An offset provider that uses the instance's class's hash code.
+         */
+        enum ForDynamicTypeHash implements OffsetProvider {
+
+            /**
+             * The singleton instance.
+             */
+            INSTANCE;
+
+            /**
+             * {@inheritDoc}
+             */
+            public StackManipulation resolve(TypeDescription instrumentedType) {
+                return new StackManipulation.Compound(MethodVariableAccess.loadThis(),
+                        MethodInvocation.invoke(GET_CLASS).virtual(instrumentedType),
+                        MethodInvocation.invoke(HASH_CODE).virtual(TypeDescription.ForLoadedType.of(Class.class)));
+            }
+        }
     }
 
     /**
@@ -274,17 +374,23 @@ public class HashCodeMethod implements Implementation {
              */
             INSTANCE;
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation before() {
                 return StackManipulation.Trivial.INSTANCE;
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation after() {
                 return StackManipulation.Trivial.INSTANCE;
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public int getRequiredVariablePadding() {
                 return StackSize.ZERO.getSize();
             }
@@ -296,16 +402,6 @@ public class HashCodeMethod implements Implementation {
         @HashCodeAndEqualsPlugin.Enhance
         class UsingJump implements NullValueGuard {
 
-            /**
-             * An empty array.
-             */
-            private static final Object[] EMPTY = new Object[0];
-
-            /**
-             * An array that only contains an integer stack map frame.
-             */
-            private static final Object[] INTEGER = new Object[]{Opcodes.INTEGER};
-
             /**
              * The instrumented method.
              */
@@ -326,17 +422,23 @@ public class HashCodeMethod implements Implementation {
                 label = new Label();
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation before() {
                 return new BeforeInstruction();
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation after() {
                 return new AfterInstruction();
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public int getRequiredVariablePadding() {
                 return 1;
             }
@@ -345,20 +447,17 @@ public class HashCodeMethod implements Implementation {
              * The stack manipulation to apply before the hash value computation.
              */
             @HashCodeAndEqualsPlugin.Enhance(includeSyntheticFields = true)
-            protected class BeforeInstruction implements StackManipulation {
+            protected class BeforeInstruction extends StackManipulation.AbstractBase {
 
-                @Override
-                public boolean isValid() {
-                    return true;
-                }
-
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
                 public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                     methodVisitor.visitVarInsn(Opcodes.ASTORE, instrumentedMethod.getStackSize());
                     methodVisitor.visitVarInsn(Opcodes.ALOAD, instrumentedMethod.getStackSize());
                     methodVisitor.visitJumpInsn(Opcodes.IFNULL, label);
                     methodVisitor.visitVarInsn(Opcodes.ALOAD, instrumentedMethod.getStackSize());
-                    return new Size(0, 0);
+                    return Size.ZERO;
                 }
             }
 
@@ -366,20 +465,17 @@ public class HashCodeMethod implements Implementation {
              * The stack manipulation to apply after the hash value computation.
              */
             @HashCodeAndEqualsPlugin.Enhance(includeSyntheticFields = true)
-            protected class AfterInstruction implements StackManipulation {
+            protected class AfterInstruction extends StackManipulation.AbstractBase {
 
-                @Override
-                public boolean isValid() {
-                    return true;
-                }
-
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
                 public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                     methodVisitor.visitLabel(label);
-                    if (implementationContext.getClassFileVersion().isAtLeast(ClassFileVersion.JAVA_V6)) {
-                        methodVisitor.visitFrame(Opcodes.F_SAME1, EMPTY.length, EMPTY, INTEGER.length, INTEGER);
-                    }
-                    return new Size(0, 0);
+                    implementationContext.getFrameGeneration().same1(methodVisitor,
+                            TypeDescription.ForLoadedType.of(int.class),
+                            Arrays.asList(implementationContext.getInstrumentedType(), TypeDescription.ForLoadedType.of(Object.class)));
+                    return Size.ZERO;
                 }
             }
         }
@@ -394,7 +490,7 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code long} value.
          */
         LONG {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitInsn(Opcodes.DUP2);
                 methodVisitor.visitIntInsn(Opcodes.BIPUSH, 32);
@@ -409,10 +505,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code float} value.
          */
         FLOAT {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Float", "floatToIntBits", "(F)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -420,7 +516,7 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code double} value.
          */
         DOUBLE {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Double", "doubleToLongBits", "(D)J", false);
                 methodVisitor.visitInsn(Opcodes.DUP2);
@@ -436,10 +532,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code boolean[]} value.
          */
         BOOLEAN_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([Z)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -447,10 +543,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code byte[]} value.
          */
         BYTE_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([B)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -458,10 +554,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code short[]} value.
          */
         SHORT_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([S)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -469,10 +565,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code char[]} value.
          */
         CHARACTER_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([C)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -480,10 +576,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for an {@code int[]} value.
          */
         INTEGER_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([I)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -491,10 +587,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code long[]} value.
          */
         LONG_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([J)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -502,10 +598,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code float[]} value.
          */
         FLOAT_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([F)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -513,10 +609,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a {@code double[]} value.
          */
         DOUBLE_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([D)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -524,10 +620,10 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a reference array value.
          */
         REFERENCE_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "hashCode", "([Ljava/lang/Object;)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
             }
         },
 
@@ -535,10 +631,21 @@ public class HashCodeMethod implements Implementation {
          * A transformer for a nested reference array value.
          */
         NESTED_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "deepHashCode", "([Ljava/lang/Object;)I", false);
-                return new Size(0, 0);
+                return Size.ZERO;
+            }
+        },
+
+        /**
+         * A transformer for computing the identity hash code for a reference.
+         */
+        REFERENCE_IDENTITY {
+            /** {@inheritDoc} */
+            public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
+                methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/System", "identityHashCode", "(Ljava/lang/Object;)I", false);
+                return Size.ZERO;
             }
         };
 
@@ -548,6 +655,7 @@ public class HashCodeMethod implements Implementation {
          * @param typeDefinition The type definition to resolve.
          * @return The stack manipulation to apply.
          */
+        @SuppressFBWarnings(value = "NP_NULL_ON_SOME_PATH_FROM_RETURN_VALUE", justification = "Assuming component type for array type.")
         public static StackManipulation of(TypeDefinition typeDefinition) {
             if (typeDefinition.represents(boolean.class)
                     || typeDefinition.represents(byte.class)
@@ -586,7 +694,9 @@ public class HashCodeMethod implements Implementation {
             }
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public boolean isValid() {
             return true;
         }
@@ -618,6 +728,11 @@ public class HashCodeMethod implements Implementation {
          */
         private final ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable;
 
+        /**
+         * A matcher to determine that a field should be considered by its identity.
+         */
+        private final ElementMatcher<? super FieldDescription.InDefinedShape> identity;
+
         /**
          * Creates a new appender for implementing a hash code method.
          *
@@ -625,18 +740,23 @@ public class HashCodeMethod implements Implementation {
          * @param multiplier        A multiplier for each value before adding a field's hash code value.
          * @param fieldDescriptions A list of fields to include in the hash code computation.
          * @param nonNullable       A matcher to determine fields of a reference type that cannot be {@code null}.
+         * @param identity          A matcher to determine that a field should be considered by its identity.
          */
         protected Appender(StackManipulation initialValue,
                            int multiplier,
                            List<FieldDescription.InDefinedShape> fieldDescriptions,
-                           ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable) {
+                           ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable,
+                           ElementMatcher<? super FieldDescription.InDefinedShape> identity) {
             this.initialValue = initialValue;
             this.multiplier = multiplier;
             this.fieldDescriptions = fieldDescriptions;
             this.nonNullable = nonNullable;
+            this.identity = identity;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Context implementationContext, MethodDescription instrumentedMethod) {
             if (instrumentedMethod.isStatic()) {
                 throw new IllegalStateException("Hash code method must not be static: " + instrumentedMethod);
@@ -651,14 +771,19 @@ public class HashCodeMethod implements Implementation {
                 stackManipulations.add(Multiplication.INTEGER);
                 stackManipulations.add(MethodVariableAccess.loadThis());
                 stackManipulations.add(FieldAccess.forField(fieldDescription).read());
-                NullValueGuard nullValueGuard = fieldDescription.getType().isPrimitive() || fieldDescription.getType().isArray() || nonNullable.matches(fieldDescription)
-                        ? NullValueGuard.NoOp.INSTANCE
-                        : new NullValueGuard.UsingJump(instrumentedMethod);
-                stackManipulations.add(nullValueGuard.before());
-                stackManipulations.add(ValueTransformer.of(fieldDescription.getType()));
-                stackManipulations.add(Addition.INTEGER);
-                stackManipulations.add(nullValueGuard.after());
-                padding = Math.max(padding, nullValueGuard.getRequiredVariablePadding());
+                if (!fieldDescription.getType().isPrimitive() && identity.matches(fieldDescription)) {
+                    stackManipulations.add(ValueTransformer.REFERENCE_IDENTITY);
+                    stackManipulations.add(Addition.INTEGER);
+                } else {
+                    NullValueGuard nullValueGuard = fieldDescription.getType().isPrimitive() || fieldDescription.getType().isArray() || nonNullable.matches(fieldDescription)
+                            ? NullValueGuard.NoOp.INSTANCE
+                            : new NullValueGuard.UsingJump(instrumentedMethod);
+                    stackManipulations.add(nullValueGuard.before());
+                    stackManipulations.add(ValueTransformer.of(fieldDescription.getType()));
+                    stackManipulations.add(Addition.INTEGER);
+                    stackManipulations.add(nullValueGuard.after());
+                    padding = Math.max(padding, nullValueGuard.getRequiredVariablePadding());
+                }
             }
             stackManipulations.add(MethodReturn.INTEGER);
             return new Size(new StackManipulation.Compound(stackManipulations).apply(methodVisitor, implementationContext).getMaximalSize(), instrumentedMethod.getStackSize() + padding);
