@@ -1,5 +1,21 @@
+/*
+ * Copyright 2014 - Present Rafael Winterhalter
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
 package net.bytebuddy.implementation.bytecode.constant;
 
+import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
 import net.bytebuddy.description.method.MethodDescription;
 import net.bytebuddy.description.type.TypeDescription;
 import net.bytebuddy.implementation.Implementation;
@@ -11,11 +27,11 @@ import net.bytebuddy.implementation.bytecode.assign.TypeCasting;
 import net.bytebuddy.implementation.bytecode.collection.ArrayFactory;
 import net.bytebuddy.implementation.bytecode.member.FieldAccess;
 import net.bytebuddy.implementation.bytecode.member.MethodInvocation;
+import net.bytebuddy.utility.nullability.MaybeNull;
 import org.objectweb.asm.MethodVisitor;
 
 import java.lang.reflect.Constructor;
 import java.lang.reflect.Method;
-import java.security.AccessController;
 import java.security.PrivilegedExceptionAction;
 import java.util.ArrayList;
 import java.util.List;
@@ -26,7 +42,39 @@ import static net.bytebuddy.matcher.ElementMatchers.isConstructor;
  * Represents the creation of a {@link java.lang.reflect.Method} value which can be created from a given
  * set of constant pool values and can therefore be considered a constant in the broader meaning.
  */
-public abstract class MethodConstant implements StackManipulation {
+public abstract class MethodConstant extends StackManipulation.AbstractBase {
+
+    /**
+     * The {@code java.security.AccessController#doPrivileged(PrivilegedExceptionAction)} method or {@code null} if
+     * this method is not available on the current VM.
+     */
+    @MaybeNull
+    protected static final MethodDescription.InDefinedShape DO_PRIVILEGED = doPrivileged();
+
+    /**
+     * Resolves the {@code java.security.AccessController#doPrivileged} method if the security manager is supported
+     * on the current VM and if security manager support is not explicitly disabled.
+     *
+     * @return The {@code doPrivileged} method or {@code null}.
+     */
+    @MaybeNull
+    @SuppressFBWarnings(value = "REC_CATCH_EXCEPTION", justification = "Exception should not be rethrown but trigger a fallback.")
+    private static MethodDescription.InDefinedShape doPrivileged() {
+        MethodDescription.InDefinedShape doPrivileged;
+        try {
+            doPrivileged = new MethodDescription.ForLoadedMethod(Class.forName("java.security.AccessController").getMethod("doPrivileged", PrivilegedExceptionAction.class));
+            try {
+                if (!Boolean.parseBoolean(System.getProperty("net.bytebuddy.securitymanager", "true"))) {
+                    doPrivileged = null;
+                }
+            } catch (SecurityException ignored) {
+                /* do nothing */
+            }
+        } catch (Exception ignored) {
+            doPrivileged = null;
+        }
+        return doPrivileged;
+    }
 
     /**
      * A description of the method to be loaded onto the stack.
@@ -60,18 +108,23 @@ public abstract class MethodConstant implements StackManipulation {
     }
 
     /**
-     * Creates a stack manipulation that loads a method constant onto the operand stack using an {@link AccessController}.
+     * Creates a stack manipulation that loads a method constant onto the operand stack using an {@code java.security.AccessController}.
+     * If the current VM does not support the access controller API, or if {@code net.bytebuddy.securitymanager} is set to false, this
+     * method has the same effect as {@link MethodConstant#of(MethodDescription.InDefinedShape)}.
      *
      * @param methodDescription The method to be loaded onto the stack.
      * @return A stack manipulation that assigns a method constant for the given method description.
      */
     public static CanCache ofPrivileged(MethodDescription.InDefinedShape methodDescription) {
+        if (DO_PRIVILEGED == null) {
+            return of(methodDescription);
+        }
         if (methodDescription.isTypeInitializer()) {
             return CanCacheIllegal.INSTANCE;
         } else if (methodDescription.isConstructor()) {
-            return new ForConstructor(methodDescription).privileged();
+            return new ForConstructor(methodDescription).withPrivilegedLookup();
         } else {
-            return new ForMethod(methodDescription).privileged();
+            return new ForMethod(methodDescription).withPrivilegedLookup();
         }
     }
 
@@ -90,28 +143,25 @@ public abstract class MethodConstant implements StackManipulation {
         return typeConstants;
     }
 
-    @Override
-    public boolean isValid() {
-        return true;
-    }
-
-    @Override
+    /**
+     * {@inheritDoc}
+     */
     public Size apply(MethodVisitor methodVisitor, Implementation.Context implementationContext) {
         return new Compound(
                 ClassConstant.of(methodDescription.getDeclaringType()),
                 methodName(),
-                ArrayFactory.forType(TypeDescription.Generic.OfNonGenericType.CLASS)
+                ArrayFactory.forType(TypeDescription.Generic.OfNonGenericType.ForLoadedType.of(Class.class))
                         .withValues(typeConstantsFor(methodDescription.getParameters().asTypeList().asErasures())),
                 MethodInvocation.invoke(accessorMethod())
         ).apply(methodVisitor, implementationContext);
     }
 
     /**
-     * Returns a method constant that uses an {@link AccessController} to look up this constant.
+     * Returns a method constant that uses an {@code java.security.AccessController} to look up this constant.
      *
-     * @return A method constant that uses an {@link AccessController} to look up this constant.
+     * @return A method constant that uses an {@code java.security.AccessController} to look up this constant.
      */
-    protected CanCache privileged() {
+    protected CanCache withPrivilegedLookup() {
         return new PrivilegedLookup(methodDescription, methodName());
     }
 
@@ -135,7 +185,7 @@ public abstract class MethodConstant implements StackManipulation {
     }
 
     @Override
-    public boolean equals(Object other) {
+    public boolean equals(@MaybeNull Object other) {
         if (this == other) {
             return true;
         } else if (other == null || getClass() != other.getClass()) {
@@ -155,17 +205,23 @@ public abstract class MethodConstant implements StackManipulation {
          */
         INSTANCE;
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public StackManipulation cached() {
             return Illegal.INSTANCE;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public boolean isValid() {
             return false;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Implementation.Context implementationContext) {
             return Illegal.INSTANCE.apply(methodVisitor, implementationContext);
         }
@@ -238,7 +294,9 @@ public abstract class MethodConstant implements StackManipulation {
                     : GET_DECLARED_METHOD;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public StackManipulation cached() {
             return new CachedMethod(this);
         }
@@ -294,33 +352,19 @@ public abstract class MethodConstant implements StackManipulation {
                     : GET_DECLARED_CONSTRUCTOR;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public StackManipulation cached() {
             return new CachedConstructor(this);
         }
     }
 
     /**
-     * Performs a privileged lookup of a method constant by using an {@link AccessController}.
+     * Performs a privileged lookup of a method constant by using an {@code java.security.AccessController}.
      */
     protected static class PrivilegedLookup implements StackManipulation, CanCache {
 
-        /**
-         * The {@link AccessController#doPrivileged(PrivilegedExceptionAction)} method.
-         */
-        private static final MethodDescription.InDefinedShape DO_PRIVILEGED;
-
-        /*
-         * Locates the access controller's do privileged method.
-         */
-        static {
-            try {
-                DO_PRIVILEGED = new MethodDescription.ForLoadedMethod(AccessController.class.getMethod("doPrivileged", PrivilegedExceptionAction.class));
-            } catch (NoSuchMethodException exception) {
-                throw new IllegalStateException("Cannot locate AccessController::doPrivileged", exception);
-            }
-        }
-
         /**
          * The method constant to load.
          */
@@ -342,20 +386,27 @@ public abstract class MethodConstant implements StackManipulation {
             this.methodName = methodName;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public boolean isValid() {
             return methodName.isValid();
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Implementation.Context implementationContext) {
+            if (DO_PRIVILEGED == null) {
+                throw new IllegalStateException("Privileged method invocation is not supported on the current VM");
+            }
             TypeDescription auxiliaryType = implementationContext.register(PrivilegedMemberLookupAction.of(methodDescription));
             return new Compound(
                     TypeCreation.of(auxiliaryType),
                     Duplication.SINGLE,
                     ClassConstant.of(methodDescription.getDeclaringType()),
                     methodName,
-                    ArrayFactory.forType(TypeDescription.Generic.OfNonGenericType.CLASS)
+                    ArrayFactory.forType(TypeDescription.Generic.OfNonGenericType.ForLoadedType.of(Class.class))
                             .withValues(typeConstantsFor(methodDescription.getParameters().asTypeList().asErasures())),
                     MethodInvocation.invoke(auxiliaryType.getDeclaredMethods().filter(isConstructor()).getOnly()),
                     MethodInvocation.invoke(DO_PRIVILEGED),
@@ -365,7 +416,9 @@ public abstract class MethodConstant implements StackManipulation {
             ).apply(methodVisitor, implementationContext);
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public StackManipulation cached() {
             return methodDescription.isConstructor()
                     ? new CachedConstructor(this)
@@ -378,7 +431,7 @@ public abstract class MethodConstant implements StackManipulation {
         }
 
         @Override
-        public boolean equals(Object other) {
+        public boolean equals(@MaybeNull Object other) {
             if (this == other) {
                 return true;
             } else if (other == null || getClass() != other.getClass()) {
@@ -413,12 +466,16 @@ public abstract class MethodConstant implements StackManipulation {
             this.methodConstant = methodConstant;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public boolean isValid() {
             return methodConstant.isValid();
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Implementation.Context implementationContext) {
             return FieldAccess.forField(implementationContext.cache(methodConstant, METHOD_TYPE))
                     .read()
@@ -431,7 +488,7 @@ public abstract class MethodConstant implements StackManipulation {
         }
 
         @Override
-        public boolean equals(Object other) {
+        public boolean equals(@MaybeNull Object other) {
             if (this == other) {
                 return true;
             } else if (other == null || getClass() != other.getClass()) {
@@ -466,12 +523,16 @@ public abstract class MethodConstant implements StackManipulation {
             this.constructorConstant = constructorConstant;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public boolean isValid() {
             return constructorConstant.isValid();
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Implementation.Context implementationContext) {
             return FieldAccess.forField(implementationContext.cache(constructorConstant, CONSTRUCTOR_TYPE))
                     .read()
@@ -484,7 +545,7 @@ public abstract class MethodConstant implements StackManipulation {
         }
 
         @Override
-        public boolean equals(Object other) {
+        public boolean equals(@MaybeNull Object other) {
             if (this == other) {
                 return true;
             } else if (other == null || getClass() != other.getClass()) {
